"use strict";
var helper = require("./helper");
var events = require("./events");
var logger = require("./logger");
var browser_result_1 = require("./browser_result");
// The browser is ready to execute tests.
var READY = 1;
// The browser is executing the tests.
var EXECUTING = 2;
// The browser is not executing, but temporarily disconnected (waiting for reconnecting).
var READY_DISCONNECTED = 3;
// The browser is executing the tests, but temporarily disconnect (waiting for reconnecting).
var EXECUTING_DISCONNECTED = 4;
// The browser got permanently disconnected (being removed from the collection and destroyed).
var DISCONNECTED = 5;
var Browser = (function () {
    function Browser(id, fullName, 
        /* capturedBrowsers */ collection, emitter, socket, timer, 
        /* config.browserDisconnectTimeout */ disconnectDelay, 
        /* config.browserNoActivityTimeout */ noActivityTimeout) {
        var _this = this;
        this.id = id;
        this.fullName = fullName;
        this.collection = collection;
        this.emitter = emitter;
        this.socket = socket;
        this.timer = timer;
        this.disconnectDelay = disconnectDelay;
        this.noActivityTimeout = noActivityTimeout;
        this.name = helper.browserFullNameToShort(this.fullName);
        this.log = logger.create(this.name);
        this.activeSockets = [this.socket];
        this.refreshNoActivityTimeout = this.noActivityTimeout ? function () {
            _this.clearNoActivityTimeout();
            _this.noActivityTimeoutId = _this.timer.setTimeout(function () {
                _this.lastResult.totalTimeEnd();
                _this.lastResult.disconnected = true;
                _this.disconnect(', because no message in ' + _this.noActivityTimeout + ' ms.');
                _this.emitter.emit('browser_complete', _this);
            }, _this.noActivityTimeout);
        } : function () {
        };
        this.clearNoActivityTimeout = this.noActivityTimeout ? function () {
            if (_this.noActivityTimeoutId) {
                _this.timer.clearTimeout(_this.noActivityTimeoutId);
                _this.noActivityTimeoutId = null;
            }
        } : function () {
        };
        this.state = READY;
        this.lastResult = new browser_result_1.Result();
        this.disconnectsCount = 0;
    }
    Browser.prototype.activeSocketsIds = function () {
        return this.activeSockets.map(function (s) {
            return s.id;
        }).join(', ');
    };
    Browser.prototype.disconnect = function (reason) {
        this.state = DISCONNECTED;
        this.disconnectsCount++;
        this.log.warn('Disconnected (%d times)' + (reason || ''), this.disconnectsCount);
        this.emitter.emit('browser_error', this, 'Disconnected' + reason);
        this.collection.remove(this);
    };
    Browser.prototype.init = function () {
        this.collection.add(this);
        events.bindAll(this, this.socket);
        this.log.info('Connected on socket %s with id %s', this.socket.id, this.id);
        // TODO(vojta): move to collection
        this.emitter.emit('browsers_change', this.collection);
        this.emitter.emit('browser_register', this);
    };
    Browser.prototype.isReady = function () {
        return this.state === READY;
    };
    Browser.prototype.toString = function () {
        return this.name;
    };
    Browser.prototype.onKarmaError = function (error) {
        if (this.isReady()) {
            return;
        }
        this.lastResult.error = true;
        this.emitter.emit('browser_error', this, error);
        this.refreshNoActivityTimeout();
    };
    Browser.prototype.onInfo = function (info) {
        if (this.isReady()) {
            return;
        }
        // TODO(vojta): remove
        if (helper.isDefined(info.dump)) {
            this.emitter.emit('browser_log', this, info.dump, 'dump');
        }
        if (helper.isDefined(info.log)) {
            this.emitter.emit('browser_log', this, info.log, info.type);
        }
        this.refreshNoActivityTimeout();
    };
    Browser.prototype.onStart = function (info) {
        this.lastResult = new browser_result_1.Result();
        this.lastResult.total = info.total;
        if (info.total === null) {
            this.log.warn('Adapter did not report total number of specs.');
        }
        this.emitter.emit('browser_start', this, info);
        this.refreshNoActivityTimeout();
    };
    Browser.prototype.onComplete = function (result) {
        if (this.isReady()) {
            return;
        }
        this.state = READY;
        this.lastResult.totalTimeEnd();
        if (!this.lastResult.success) {
            this.lastResult.error = true;
        }
        this.emitter.emit('browsers_change', this.collection);
        this.emitter.emit('browser_complete', this, result);
        this.clearNoActivityTimeout();
    };
    Browser.prototype.onDisconnect = function (_, disconnectedSocket) {
        var _this = this;
        this.activeSockets.splice(this.activeSockets.indexOf(disconnectedSocket), 1);
        if (this.activeSockets.length) {
            this.log.debug('Disconnected %s, still have %s', disconnectedSocket.id, this.activeSocketsIds());
            return;
        }
        if (this.state === READY) {
            this.disconnect();
        }
        else if (this.state === EXECUTING) {
            this.log.debug('Disconnected during run, waiting %sms for reconnecting.', this.disconnectDelay);
            this.state = EXECUTING_DISCONNECTED;
            this.pendingDisconnect = this.timer.setTimeout(function () {
                _this.lastResult.totalTimeEnd();
                _this.lastResult.disconnected = true;
                _this.disconnect();
                _this.emitter.emit('browser_complete', _this);
            }, this.disconnectDelay);
            this.clearNoActivityTimeout();
        }
    };
    Browser.prototype.reconnect = function (newSocket) {
        if (this.state === EXECUTING_DISCONNECTED) {
            this.state = EXECUTING;
            this.log.debug('Reconnected on %s.', newSocket.id);
        }
        else if (this.state === EXECUTING || this.state === READY) {
            this.log.debug('New connection %s (already have %s)', newSocket.id, this.activeSocketsIds());
        }
        else if (this.state === DISCONNECTED) {
            this.state = READY;
            this.log.info('Connected on socket %s with id %s', newSocket.id, this.id);
            this.collection.add(this);
            // TODO(vojta): move to collection
            this.emitter.emit('browsers_change', this.collection);
            this.emitter.emit('browser_register', this);
        }
        var exists = this.activeSockets.some(function (s) { return s.id === newSocket.id; });
        if (!exists) {
            this.activeSockets.push(newSocket);
            events.bindAll(this, newSocket);
        }
        if (this.pendingDisconnect) {
            this.timer.clearTimeout(this.pendingDisconnect);
        }
        this.refreshNoActivityTimeout();
    };
    Browser.prototype.onResult = function (result) {
        if (result.length) {
            return result.forEach(this.onResult, this);
        }
        // ignore - probably results from last run (after server disconnecting)
        if (this.isReady()) {
            return;
        }
        this.lastResult.add(result);
        this.emitter.emit('spec_complete', this, result);
        this.refreshNoActivityTimeout();
    };
    Browser.prototype.serialize = function () {
        return {
            id: this.id,
            name: this.name,
            isReady: this.state === READY
        };
    };
    Browser.prototype.execute = function (config) {
        this.activeSockets.forEach(function (socket) {
            socket.emit('execute', config);
        });
        this.state = EXECUTING;
        this.refreshNoActivityTimeout();
    };
    return Browser;
}());
exports.Browser = Browser;
Browser.STATE_READY = READY;
Browser.STATE_EXECUTING = EXECUTING;
Browser.STATE_READY_DISCONNECTED = READY_DISCONNECTED;
Browser.STATE_EXECUTING_DISCONNECTED = EXECUTING_DISCONNECTED;
Browser.STATE_DISCONNECTED = DISCONNECTED;
//# sourceMappingURL=browser.js.map