"use strict";
var Promise = require("bluebird");
var Jobs = require('qjobs');
var helper = require("./helper");
var logger_1 = require("./logger");
var log = logger_1.create('launcher');
var base_1 = require("./launchers/base");
var capture_timeout_1 = require("./launchers/capture_timeout");
var process_1 = require("./launchers/process");
var retry_1 = require("./launchers/retry");
var baseDecorator = base_1.BaseLauncher.decoratorFactory;
var captureTimeoutDecorator = capture_timeout_1.CaptureTimeoutLauncher.decoratorFactory;
var retryDecorator = retry_1.RetryLauncher.decoratorFactory;
var processDecorator = process_1.ProcessLauncher.decoratorFactory;
// TODO(vojta): remove once nobody uses it
function baseBrowserDecoratorFactory(baseLauncherDecorator, captureTimeoutLauncherDecorator, retryLauncherDecorator, processLauncherDecorator) {
    return function (launcher) {
        baseLauncherDecorator(launcher);
        captureTimeoutLauncherDecorator(launcher);
        retryLauncherDecorator(launcher);
        processLauncherDecorator(launcher);
    };
}
var Launcher = (function () {
    function Launcher(server, emitter, injector) {
        var _this = this;
        this.server = server;
        this.emitter = emitter;
        this.injector = injector;
        this._browsers = [];
        this.killAll = function (callback) {
            log.debug('Disconnecting all browsers');
            var remaining = 0;
            var finish = function () {
                remaining--;
                if (!remaining && callback) {
                    callback();
                }
            };
            if (!_this._browsers.length) {
                return process.nextTick(callback);
            }
            _this._browsers.forEach(function (browser) {
                remaining++;
                browser.forceKill().then(finish);
            });
        };
        this.launch.$inject = [
            'config.browsers',
            'config.concurrency'
        ];
        this.launchSingle.$inject = [
            'config.protocol',
            'config.hostname',
            'config.port',
            'config.urlRoot',
            'config.upstreamProxy'
        ];
        // register events
        emitter.on('exit', this.killAll);
    }
    Launcher.prototype.getBrowserById = function (id) {
        // return this._browsers.find(browser => browser.id === id)
        for (var i = 0; i < this._browsers.length; i++) {
            if (this._browsers[i].id === id) {
                return this._browsers[i];
            }
        }
        return null;
    };
    Launcher.prototype.launchSingle = function (protocol, hostname, port, urlRoot, upstreamProxy) {
        var _this = this;
        return function (name) {
            if (upstreamProxy) {
                protocol = upstreamProxy.protocol;
                hostname = upstreamProxy.hostname;
                port = upstreamProxy.port;
                urlRoot = upstreamProxy.path + urlRoot.substr(1);
            }
            var url = protocol + '//' + hostname + ':' + port + urlRoot;
            var locals = {
                id: ['value', Launcher.generateId()],
                name: ['value', name],
                baseLauncherDecorator: ['factory', baseDecorator],
                captureTimeoutLauncherDecorator: ['factory', captureTimeoutDecorator],
                retryLauncherDecorator: ['factory', retryDecorator],
                processLauncherDecorator: ['factory', processDecorator],
                baseBrowserDecorator: ['factory', baseBrowserDecoratorFactory]
            };
            // TODO(vojta): determine script from name
            if (name.indexOf('/') !== -1) {
                name = 'Script';
            }
            try {
                var browser = _this.injector.createChild([locals], ['launcher:' + name]).get('launcher:' + name);
            }
            catch (e) {
                if (e.message.indexOf('No provider for "launcher:' + name + '"') !== -1) {
                    log.error('Cannot load browser "%s": it is not registered! ' +
                        'Perhaps you are missing some plugin?', name);
                }
                else {
                    log.error('Cannot load browser "%s"!\n  ' + e.stack, name);
                }
                _this.emitter.emit('load_error', 'launcher', name);
                return;
            }
            // TODO(vojta): remove in v1.0 (BC for old launchers)
            if (!browser.forceKill) {
                browser.forceKill = function () {
                    var me = this;
                    return new Promise(function (resolve) {
                        me.kill(resolve);
                    });
                };
                browser.restart = function () {
                    var me = this;
                    this.kill(function () {
                        me.start(url);
                    });
                };
            }
            _this.jobs.add(function (args, done) {
                log.info('Starting browser %s', helper.isDefined(browser.displayName) ? browser.displayName : browser.name);
                browser.on('browser_process_failure', function () {
                    done(browser.error);
                });
                browser.on('done', function () {
                    // We are not done if there was an error as first the retry takes
                    // place which we catch with `browser_process_failure` if it fails
                    if (browser.error || browser.state === browser.STATE_RESTARTING)
                        return;
                    done(null, browser);
                });
                browser.start(url);
            }, []);
            _this.jobs.run();
            _this._browsers.push(browser);
        };
    };
    Launcher.prototype.launch = function (names, concurrency) {
        var _this = this;
        log.info('Launching browser%s %s with %s', names.length > 1 ? 's' : '', names.join(', '), concurrency === Infinity ? 'unlimited concurrency' : 'concurrency ' + concurrency);
        this.jobs = new Jobs({ maxConcurrency: concurrency });
        this.lastStartTime = Date.now();
        if (this.server.loadErrors.length === 0) {
            names.forEach(function (name) {
                _this.injector.invoke(_this.launchSingle, _this)(name);
            });
        }
        else {
            // Empty task to ensure `end` is emitted
            this.jobs.add(function (args, done) {
                done();
            }, []);
        }
        this.jobs.on('end', function (err) {
            log.debug('Finished all browsers');
            if (err) {
                log.error(err);
            }
        });
        this.jobs.run();
        return this._browsers;
    };
    Launcher.prototype.kill = function (id, callback) {
        var browser = this.getBrowserById(id);
        callback = callback || function () {
        };
        if (!browser) {
            process.nextTick(callback);
            return false;
        }
        browser.forceKill().then(callback);
        return true;
    };
    Launcher.prototype.restart = function (id) {
        var browser = this.getBrowserById(id);
        if (!browser) {
            return false;
        }
        browser.restart();
        return true;
    };
    Launcher.prototype.areAllCaptured = function () {
        return !this._browsers.some(function (browser) {
            return !browser.isCaptured();
        });
    };
    Launcher.prototype.markCaptured = function (id) {
        var _this = this;
        this._browsers.forEach(function (browser) {
            if (browser.id === id) {
                browser.markCaptured();
                log.debug('%s (id %s) captured in %d secs', browser.name, browser.id, (Date.now() - _this.lastStartTime) / 1000);
            }
        });
    };
    Launcher.generateId = function () {
        return '' + Math.floor(Math.random() * 100000000);
    };
    return Launcher;
}());
exports.Launcher = Launcher;
Launcher.$inject = ['server', 'emitter', 'injector'];
//# sourceMappingURL=launcher.js.map