"use strict";
var path = require("path");
var log = require('../logger').create('launcher');
var env = process.env;
var ProcessLauncher = (function () {
    function ProcessLauncher(spawn, tempDir, timer) {
        var _this = this;
        this.spawn = spawn;
        this.tempDir = tempDir;
        this.timer = timer;
        this.killTimeout = 2000;
        this._start = function (url) {
            _this._execCommand(_this._getCommand(), _this._getOptions(url));
        };
        this._getCommand = function () {
            return env[_this.ENV_CMD] || _this.DEFAULT_CMD[process.platform];
        };
        this._getOptions = function (url) {
            return [url];
        };
        // Normalize the command, remove quotes (spawn does not like them).
        this._normalizeCommand = function (cmd) {
            if (cmd.charAt(0) === cmd.charAt(cmd.length - 1) && '\'`"'.indexOf(cmd.charAt(0)) !== -1) {
                cmd = cmd.substring(1, cmd.length - 1);
                log.warn('The path should not be quoted.\n  Normalized the path to %s', cmd);
            }
            return path.normalize(cmd);
        };
        this._execCommand = function (cmd, args) {
            if (!cmd) {
                log.error('No binary for %s browser on your platform.\n  ' +
                    'Please, set "%s" env variable.', _this.name, _this.ENV_CMD);
                // disable restarting
                _this._retryLimit = -1;
                return _this._clearTempDirAndReportDone('no binary');
            }
            cmd = _this._normalizeCommand(cmd);
            log.debug(cmd + ' ' + args.join(' '));
            _this._process = _this.spawn(cmd, args);
            var errorOutput = '';
            _this._process.on('exit', function (code) {
                _this._onProcessExit(code, errorOutput);
            });
            _this._process.on('error', function (err) {
                if (err.code === 'ENOENT') {
                    _this._retryLimit = -1;
                    errorOutput = 'Can not find the binary ' + cmd + '\n\t' +
                        'Please set env variable ' + _this.ENV_CMD;
                }
                else {
                    errorOutput += err.toString();
                }
            });
        };
        this._onProcessExit = function (code, errorOutput) {
            log.debug('Process %s exited with code %d', _this.name, code);
            var error = null;
            if (_this.state === _this.STATE_BEING_CAPTURED) {
                log.error('Cannot start %s\n\t%s', _this.name, errorOutput);
                error = 'cannot start';
            }
            if (_this.state === _this.STATE_CAPTURED) {
                log.error('%s crashed.\n\t%s', _this.name, errorOutput);
                error = 'crashed';
            }
            _this.process = null;
            if (_this._killTimer) {
                _this.timer.clearTimeout(_this._killTimer);
                _this._killTimer = null;
            }
            _this._clearTempDirAndReportDone(error);
        };
        this._clearTempDirAndReportDone = function (error) {
            _this.tempDir.remove(_this._tempDir, function () {
                _this._done(error);
                if (_this.onExitCallback) {
                    _this.onExitCallback();
                    _this.onExitCallback = null;
                }
            });
        };
        this._onKillTimeout = function () {
            if (_this.state !== _this.STATE_BEING_KILLED && _this.state !== _this.STATE_BEING_FORCE_KILLED) {
                return;
            }
            log.warn('%s was not killed in %d ms, sending SIGKILL.', _this.name, _this.killTimeout);
            _this._process.kill('SIGKILL');
            // NOTE: https://github.com/karma-runner/karma/pull/1184
            // NOTE: SIGKILL is just a signal.  Processes should never ignore it, but they can.
            // If a process gets into a state where it doesn't respond in a reasonable amount of time
            // Karma should warn, and continue as though the kill succeeded.
            // This a certainly suboptimal, but it is better than having the test harness hang waiting
            // for a zombie child process to exit.
            _this._killTimer = _this.timer.setTimeout(function () {
                log.warn('%s was not killed by SIGKILL in %d ms, continuing.', _this.name, _this.killTimeout);
                _this._onProcessExit(-1, '');
            }, _this.killTimeout);
        };
        this._tempDir = tempDir.getPath('/karma-' + this.id.toString());
        this.on('start', function (url) {
            tempDir.create(_this._tempDir);
            _this._start(url);
        });
        this.on('kill', function (done) {
            if (!_this._process) {
                return process.nextTick(done);
            }
            _this.onExitCallback = done;
            _this._process.kill();
            _this._killTimer = timer.setTimeout(_this._onKillTimeout, _this.killTimeout);
        });
    }
    ProcessLauncher.decoratorFactory = function (timer) {
        return function (launcher) {
            var spawn = require('child_process').spawn;
            var spawnWithoutOutput = function () {
                var proc = spawn.apply(null, arguments);
                proc.stdout.resume();
                proc.stderr.resume();
                return proc;
            };
            ProcessLauncher.call(launcher, spawnWithoutOutput, require('../temp_dir'), timer);
        };
    };
    return ProcessLauncher;
}());
exports.ProcessLauncher = ProcessLauncher;
//# sourceMappingURL=process.js.map