"use strict";
var util = require("util");
var helper = require("../helper");
var BaseReporter = (function () {
    function BaseReporter(formatError, reportSlow, useColors, browserConsoleLogOptions, adapter) {
        var _this = this;
        this.formatError = formatError;
        this.reportSlow = reportSlow;
        this.useColors = useColors;
        this.browserConsoleLogOptions = browserConsoleLogOptions;
        this.onRunStart = function () {
            _this._browsers = [];
        };
        this.onBrowserStart = function (browser) {
            _this._browsers.push(browser);
        };
        this.renderBrowser = function (browser) {
            var results = browser.lastResult;
            var totalExecuted = results.success + results.failed;
            var msg = util.format('%s: Executed %d of %d', browser, totalExecuted, results.total);
            if (results.failed) {
                msg += util.format(_this.X_FAILED, results.failed);
            }
            if (results.skipped) {
                msg += util.format(' (skipped %d)', results.skipped);
            }
            if (browser.isReady) {
                if (results.disconnected) {
                    msg += _this.FINISHED_DISCONNECTED;
                }
                else if (results.error) {
                    msg += _this.FINISHED_ERROR;
                }
                else if (!results.failed) {
                    msg += _this.FINISHED_SUCCESS;
                }
                msg += util.format(' (%s / %s)', helper.formatTimeInterval(results.totalTime), helper.formatTimeInterval(results.netTime));
            }
            return msg;
        };
        this.write = function () {
            var _arguments = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                _arguments[_i - 0] = arguments[_i];
            }
            var msg = util.format.apply(null, Array.prototype.slice.call(_arguments));
            var self = _this;
            _this.adapters.forEach(function (adapter) {
                if (!helper.isDefined(adapter.colors)) {
                    adapter.colors = _this.useColors;
                }
                if (!helper.isDefined(self.EXCLUSIVELY_USE_COLORS) || adapter.colors === self.EXCLUSIVELY_USE_COLORS) {
                    return adapter(msg);
                }
            });
        };
        this.writeCommonMsg = this.write;
        this.onBrowserError = function (browser, error) {
            _this.writeCommonMsg(util.format(_this.ERROR, browser) + _this.formatError(error, '  '));
        };
        this.onBrowserLog = function (browser, log, type) {
            if (!_this.browserConsoleLogOptions || !_this.browserConsoleLogOptions.terminal)
                return;
            if (!helper.isString(log)) {
                // TODO(vojta): change util to new syntax (config object)
                log = util.inspect(log, false, undefined, _this.USE_COLORS);
            }
            if (_this._browsers && _this._browsers.length === 1) {
                _this.writeCommonMsg(util.format(_this.LOG_SINGLE_BROWSER, type.toUpperCase(), log));
            }
            else {
                _this.writeCommonMsg(util.format(_this.LOG_MULTI_BROWSER, browser, type.toUpperCase(), log));
            }
        };
        this.onSpecComplete = function (browser, result) {
            if (result.skipped) {
                _this.specSkipped(browser, result);
            }
            else if (result.success) {
                _this.specSuccess(browser, result);
            }
            else {
                _this.specFailure(browser, result);
            }
            if (_this.reportSlow && result.time > _this.reportSlow) {
                var specName = result.suite.join(' ') + ' ' + result.description;
                var time = helper.formatTimeInterval(result.time);
                _this.writeCommonMsg(util.format(_this.SPEC_SLOW, browser, time, specName));
            }
        };
        this.specSkipped = function (browser, result) {
        };
        this.specSuccess = this.specSkipped;
        this.specFailure = function (browser, result) {
            var specName = result.suite.join(' ') + ' ' + result.description;
            var msg = util.format(_this.SPEC_FAILURE, browser, specName);
            result.log.forEach(function (log) {
                msg += _this.formatError(log, '\t');
            });
            _this.writeCommonMsg(msg);
        };
        this.onRunComplete = function (browsers, results) {
            if (browsers.length > 1 && !results.error && !results.disconnected) {
                if (!results.failed) {
                    _this.write(_this.TOTAL_SUCCESS, results.success);
                }
                else {
                    _this.write(_this.TOTAL_FAILED, results.failed, results.success);
                }
            }
        };
        this.USE_COLORS = false;
        this.EXCLUSIVELY_USE_COLORS = undefined;
        this.LOG_SINGLE_BROWSER = '%s: %s\n';
        this.LOG_MULTI_BROWSER = '%s %s: %s\n';
        this.SPEC_FAILURE = '%s %s FAILED' + '\n';
        this.SPEC_SLOW = '%s SLOW %s: %s\n';
        this.ERROR = '%s ERROR\n';
        this.FINISHED_ERROR = ' ERROR';
        this.FINISHED_SUCCESS = ' SUCCESS';
        this.FINISHED_DISCONNECTED = ' DISCONNECTED';
        this.X_FAILED = ' (%d FAILED)';
        this.TOTAL_SUCCESS = 'TOTAL: %d SUCCESS\n';
        this.TOTAL_FAILED = 'TOTAL: %d FAILED, %d SUCCESS\n';
        this.adapters = [adapter || process.stdout.write.bind(process.stdout)];
        this.renderBrowser = this.renderBrowser.bind(this);
    }
    BaseReporter.decoratorFactory = function (formatError, reportSlow, useColors, browserConsoleLogOptions) {
        return function (self) { return BaseReporter.call(self, formatError, reportSlow, useColors, browserConsoleLogOptions); };
    };
    return BaseReporter;
}());
exports.BaseReporter = BaseReporter;
BaseReporter.decoratorFactory.$inject = [
    'formatError',
    'config.reportSlowerThan',
    'config.colors',
    'config.browserConsoleLogOptions'
];
//# sourceMappingURL=base.js.map